package dev.pfaff.altargui;

import java.util.List;

/**
 * Responsible for including children for input events and painting.
 */
public sealed interface MountContext permits GuiContext.MountContextImpl {
	void pushScissor(float x, float y, float width, float height);

	default void pushScissor(WidgetInstance element) {
		pushScissor(element.clippedX(),
					element.clippedY(),
					element.clippedWidth(),
					element.clippedHeight());
	}

	void popScissor();

	void iSpecMount(WidgetInstance<?> element);

	/**
	 * Mounts a child. The child must have been laid out before calling this method.
	 */
	void mount(WidgetInstance<?> element);

	/**
	 * @deprecated {@link WidgetInstanceList} offers improved performance over generic {@link List}.
	 */
	@Deprecated
	default void mountList(List<WidgetInstance> elements) {
		elements.forEach(this::mount);
	}

	default void mountList(WidgetInstanceList elements) {
		mountArray(elements.elements(), elements.size());
	}

	default void mountArray(WidgetInstance[] elements) {
		mountArray(elements, elements.length);
	}

	default void mountArray(WidgetInstance[] elements, int count) {
		if (count > elements.length) throw new AssertionError();
		for (int i = 0; i < count; i++) {
			mount(elements[i]);
		}
	}
}
