package dev.pfaff.altargui.pathfinder;

import java.lang.foreign.Arena;
import java.lang.foreign.MemorySegment;
import java.lang.foreign.ValueLayout;

import static dev.pfaff.altargui.pathfinder.Pathfinder.*;

public record PFCanvasRef(MemorySegment ptr) {
	public PFSceneRef createScene() {
		try {
			return new PFSceneRef((MemorySegment) MH_PFCanvasCreateScene.invokeExact(this.ptr()));
		} catch (Throwable e) {
			throw new RuntimeException(e);
		}
	}

	public void setFont(String name) {
		try (var arena = Arena.ofConfined()) {
			var nameBytes = arena.allocateFrom(name);
			int result = (int) MH_PFCanvasSetFontByPostScriptName.invokeExact(this.ptr(), nameBytes, nameBytes.byteSize() - 1);
			if (result != 0) throw new RuntimeException(STR."Failed to set font to '\{name}': \{result}");
		} catch (Throwable e) {
			throw new RuntimeException(e);
		}
	}

	public void setFontSize(float size) {
		try {
			MH_PFCanvasSetFontSize.invokeExact(this.ptr(), size);
		} catch (Throwable e) {
			throw new RuntimeException(e);
		}
	}

	public void fillText(String text, float x, float y) {
		try (var arena = Arena.ofConfined()) {
			var textBytes = arena.allocateFrom(text);
			var position = arena.allocateFrom(ValueLayout.JAVA_FLOAT, x, y).asSlice(0, ML_PFVector2F);
			MH_PFCanvasFillText.invokeExact(this.ptr(), textBytes, textBytes.byteSize() - 1, position);
		} catch (Throwable e) {
			throw new RuntimeException(e);
		}
	}

	public void fillRect(float x1, float y1, float x2, float y2) {
		try (var arena = Arena.ofConfined()) {
			var rect = arena.allocateFrom(ValueLayout.JAVA_FLOAT, x1, y1, x2, y2).asSlice(0, ML_PFRectF);
			MH_PFCanvasFillRect.invokeExact(this.ptr(), rect);
		} catch (Throwable e) {
			throw new RuntimeException(e);
		}
	}

	public void setFillStyle(PFFillStyleRef style) {
		try {
			MH_PFCanvasSetFillStyle.invokeExact(this.ptr(), style.ptr());
		} catch (Throwable e) {
			throw new RuntimeException(e);
		}
	}

	public void setStrokeStyle(PFFillStyleRef style) {
		try {
			MH_PFCanvasSetStrokeStyle.invokeExact(this.ptr(), style.ptr());
		} catch (Throwable e) {
			throw new RuntimeException(e);
		}
	}

	public void clear() {
		try {
			MH_PFCanvasClear.invokeExact(this.ptr());
		} catch (Throwable e) {
			throw new RuntimeException(e);
		}
	}
}
