package dev.pfaff.altargui.pathfinder;

import java.lang.foreign.*;
import java.lang.invoke.MethodHandle;
import java.lang.invoke.MethodHandles;
import java.nio.charset.StandardCharsets;
import java.nio.file.Path;

import static java.lang.foreign.MemoryLayout.PathElement.groupElement;
import static java.lang.foreign.ValueLayout.*;
import static java.lang.invoke.MethodHandles.dropArguments;
import static java.lang.invoke.MethodHandles.explicitCastArguments;
import static java.lang.invoke.MethodType.methodType;

public final class Pathfinder {
	private static final Linker LINKER = Linker.nativeLinker();
	private static final SymbolLookup LOOKUP = SymbolLookup.libraryLookup(Path.of("natives/libpathfinder.dylib"),
			Arena.ofAuto());
	private static final MethodHandle MH_lookupGLFunction;
	private static final FunctionDescriptor FD_lookupGLFunction = FunctionDescriptor.of(ADDRESS,
			ADDRESS,
			ADDRESS);
	public static final FunctionDescriptor FD_PFGLLoadWith = FunctionDescriptor.ofVoid(ADDRESS,
			ADDRESS);
	public static StructLayout ML_PFVector2I = MemoryLayout.structLayout(ValueLayout.JAVA_INT, ValueLayout.JAVA_INT);
	public static StructLayout ML_PFVector2F = MemoryLayout.structLayout(ValueLayout.JAVA_FLOAT,
			ValueLayout.JAVA_FLOAT);
	public static StructLayout ML_PFRectF = MemoryLayout.structLayout(ML_PFVector2F, ML_PFVector2F);
	public static AddressLayout ML_PFDestFramebufferRef = ADDRESS.withName("PFDestFramebufferRef");
	public static MemoryLayout ML_PFGLRendererRef = ADDRESS.withName("PFGLRendererRef");
	public static MemoryLayout ML_PFCanvasRef = ADDRESS.withName("PFCanvasRef");
	public static MemoryLayout ML_PFFillStyleRef = ADDRESS.withName("PFFillStyleRef");
	public static MemoryLayout ML_PFSceneRef = ADDRESS.withName("PFSceneRef");
	public static MemoryLayout ML_PFSceneProxyRef = ADDRESS.withName("PFSceneProxyRef");
	public static MemoryLayout ML_PFCanvasFontContextRef = ADDRESS.withName("PFCanvasFontContextRef");
	public static MemoryLayout ML_PFColorF = MemoryLayout.structLayout(ValueLayout.JAVA_FLOAT.withName("r"),
			ValueLayout.JAVA_FLOAT.withName("g"),
			ValueLayout.JAVA_FLOAT.withName("b"),
			ValueLayout.JAVA_FLOAT.withName("a"));
	public static MemoryLayout ML_PFColorU = MemoryLayout.structLayout(ValueLayout.JAVA_BYTE.withName("r"),
			ValueLayout.JAVA_BYTE.withName("g"),
			ValueLayout.JAVA_BYTE.withName("b"),
			ValueLayout.JAVA_BYTE.withName("a"));
	public static final ValueLayout.OfByte ML_PFRendererMode = JAVA_BYTE.withName("PFRendererMode");
	public static StructLayout ML_PFRendererOptions = MemoryLayout.structLayout(ML_PFDestFramebufferRef,
			ML_PFColorF,
			ValueLayout.JAVA_BYTE);
	public static MethodHandle MH_PFGLDestFramebufferCreateFullWindow = LINKER.downcallHandle(LOOKUP.find(
					"PFGLDestFramebufferCreateFullWindow").get(),
			FunctionDescriptor.of(ML_PFDestFramebufferRef, ADDRESS.withTargetLayout(ML_PFVector2I)));
	public static MethodHandle MH_PFGLDeviceCreate = LINKER.downcallHandle(LOOKUP.find(
					"PFGLDeviceCreate").get(),
			FunctionDescriptor.of(ADDRESS, ValueLayout.JAVA_BYTE.withName("PFGLVersion"), ValueLayout.JAVA_INT));
	public static MethodHandle MH_PFEmbeddedResourceLoaderCreate = LINKER.downcallHandle(LOOKUP.find(
					"PFEmbeddedResourceLoaderCreate").get(),
			FunctionDescriptor.of(ADDRESS));
	public static MethodHandle MH_PFFilesystemResourceLoaderLocate = LINKER.downcallHandle(LOOKUP.find(
					"PFFilesystemResourceLoaderLocate").get(),
			FunctionDescriptor.of(ADDRESS));
	public static MethodHandle MH_PFFilesystemResourceLoaderFromPath = LINKER.downcallHandle(LOOKUP.find(
					"PFFilesystemResourceLoaderFromPath").get(),
			FunctionDescriptor.of(ADDRESS, ADDRESS.withTargetLayout(JAVA_BYTE)));
	public static MethodHandle MH_PFGLRendererCreate = LINKER.downcallHandle(LOOKUP.find(
					"PFGLRendererCreate").get(),
			FunctionDescriptor.of(ML_PFGLRendererRef,
					ADDRESS,
					ADDRESS,
					ADDRESS.withTargetLayout(ML_PFRendererMode),
					ADDRESS.withTargetLayout(ML_PFRendererOptions)));
	public static MethodHandle MH_PFCanvasFontContextCreateWithSystemSource = LINKER.downcallHandle(LOOKUP.find(
					"PFCanvasFontContextCreateWithSystemSource").get(),
			FunctionDescriptor.of(ML_PFCanvasFontContextRef));
	public static MethodHandle MH_PFCanvasCreate = LINKER.downcallHandle(LOOKUP.find(
					"PFCanvasCreate").get(),
			FunctionDescriptor.of(ML_PFCanvasRef,
					ML_PFCanvasFontContextRef,
					ADDRESS.withTargetLayout(ML_PFVector2F)));
	public static MethodHandle MH_PFCanvasSetFontByPostScriptName;

	static {
		var MH = LINKER.downcallHandle(LOOKUP.find(
						"PFCanvasSetFontByPostScriptName").get(),
				FunctionDescriptor.of(JAVA_INT,
						ML_PFCanvasRef,
						ADDRESS.withTargetLayout(JAVA_BYTE),
						ADDRESS.byteSize() == 8 ? JAVA_LONG : JAVA_INT));
		MH_PFCanvasSetFontByPostScriptName = explicitCastArguments(MH, MH.type().changeParameterType(2, long.class));
	}

	public static MethodHandle MH_PFCanvasSetFontSize = LINKER.downcallHandle(LOOKUP.find(
					"PFCanvasSetFontSize").get(),
			FunctionDescriptor.ofVoid(ML_PFCanvasRef,
					JAVA_FLOAT));

	public static MethodHandle MH_PFCanvasFillText = LINKER.downcallHandle(LOOKUP.find(
					"PFCanvasFillText").get(),
			FunctionDescriptor.ofVoid(ML_PFCanvasRef,
					ADDRESS.withTargetLayout(JAVA_BYTE),
					ADDRESS.byteSize() == 8 ? JAVA_LONG : JAVA_INT,
					ADDRESS.withTargetLayout(ML_PFVector2F)));
	public static MethodHandle MH_PFCanvasFillRect = LINKER.downcallHandle(LOOKUP.find(
					"PFCanvasFillRect").get(),
			FunctionDescriptor.ofVoid(ML_PFCanvasRef, ADDRESS.withTargetLayout(ML_PFRectF)));
	public static MethodHandle MH_PFCanvasSetFillStyle = LINKER.downcallHandle(LOOKUP.find(
					"PFCanvasSetFillStyle").get(),
			FunctionDescriptor.ofVoid(ML_PFCanvasRef, ML_PFFillStyleRef));
	public static MethodHandle MH_PFCanvasSetStrokeStyle = LINKER.downcallHandle(LOOKUP.find(
					"PFCanvasSetStrokeStyle").get(),
			FunctionDescriptor.ofVoid(ML_PFCanvasRef, ML_PFFillStyleRef));
	public static MethodHandle MH_PFCanvasClear = LINKER.downcallHandle(LOOKUP.find(
					"PFCanvasClear").get(),
			FunctionDescriptor.ofVoid(ML_PFCanvasRef));

	public static MethodHandle MH_PFCanvasCreateScene = LINKER.downcallHandle(LOOKUP.find(
					"PFCanvasCreateScene").get(),
			FunctionDescriptor.of(ML_PFSceneRef, ML_PFCanvasRef));
	public static MethodHandle MH_PFSceneProxyCreateFromSceneAndSequentialExecutor = LINKER.downcallHandle(LOOKUP.find(
					"PFSceneProxyCreateFromSceneAndSequentialExecutor").get(),
			FunctionDescriptor.of(ML_PFSceneProxyRef, ML_PFSceneRef, ML_PFRendererMode));
	public static MethodHandle MH_PFSceneProxyCreateFromSceneAndRayonExecutor = LINKER.downcallHandle(LOOKUP.find(
					"PFSceneProxyCreateFromSceneAndRayonExecutor").get(),
			FunctionDescriptor.of(ML_PFSceneProxyRef, ML_PFSceneRef, ML_PFRendererMode));
	public static MethodHandle MH_PFSceneProxyDestroy = LINKER.downcallHandle(LOOKUP.find(
					"PFSceneProxyDestroy").get(),
			FunctionDescriptor.ofVoid(ML_PFSceneProxyRef));
	private static final MemoryLayout ML_PFBuildOptionsRef = ADDRESS.withName("PFBuildOptions");
	public static MethodHandle MH_PFSceneProxyBuildAndRenderGL = LINKER.downcallHandle(LOOKUP.find(
					"PFSceneProxyBuildAndRenderGL").get(),
			FunctionDescriptor.ofVoid(ML_PFSceneProxyRef, ML_PFGLRendererRef, ML_PFBuildOptionsRef));
	public static MethodHandle MH_PFBuildOptionsCreate = LINKER.downcallHandle(LOOKUP.find(
					"PFBuildOptionsCreate").get(),
			FunctionDescriptor.of(ML_PFBuildOptionsRef));

	public static MethodHandle MH_PFFillStyleCreateColor = LINKER.downcallHandle(LOOKUP.find(
					"PFFillStyleCreateColor").get(),
			FunctionDescriptor.of(ML_PFFillStyleRef, ADDRESS.withTargetLayout(ML_PFColorU)));
	public static MethodHandle MH_PFFillStyleDestroy = LINKER.downcallHandle(LOOKUP.find(
					"PFFillStyleDestroy").get(),
			FunctionDescriptor.ofVoid(ML_PFFillStyleRef));

	static {
		var l = MethodHandles.lookup();
		try {
			MH_lookupGLFunction = dropArguments(l.findStatic(Pathfinder.class,
					"lookupGLFunction",
					methodType(MemorySegment.class, SymbolLookup.class, MemorySegment.class)), 2, MemorySegment.class);
		} catch (ReflectiveOperationException e) {
			throw new RuntimeException(e);
		}

		try {
			LINKER.downcallHandle(FunctionDescriptor.ofVoid())
					.invokeExact(LOOKUP.find("init_logging").get());
		} catch (Throwable e) {
			throw new RuntimeException(e);
		}
	}

	private static MemorySegment lookupGLFunction(SymbolLookup symbolLookup, MemorySegment namePtr) {
		// set a reasonable max length, because the segment we get has length 0
		var name = namePtr.reinterpret(128).getString(0, StandardCharsets.US_ASCII);
		return symbolLookup.find(name).orElse(MemorySegment.NULL);
	}

	public Pathfinder(SymbolLookup symbolLookup) {
		try (var arena = Arena.ofConfined()) {
			var glContextStub = LINKER.upcallStub(MH_lookupGLFunction.bindTo(symbolLookup), FD_lookupGLFunction, arena);
			var pfglLoadWith = LINKER.downcallHandle(FD_PFGLLoadWith);
			pfglLoadWith.invokeExact(LOOKUP.find("PFGLLoadWith").get(), glContextStub, MemorySegment.NULL);
		} catch (Throwable e) {
			throw new RuntimeException(e);
		}
	}

	public PFGLDestFramebufferRef createFullWindowFramebufferRef(int windowWidth, int windowHeight) {
		try (var arena = Arena.ofConfined()) {
			var size = arena.allocateFrom(JAVA_INT, windowWidth, windowHeight)
					.asSlice(0, ML_PFVector2I);
			return new PFGLDestFramebufferRef((MemorySegment) MH_PFGLDestFramebufferCreateFullWindow.invokeExact(size));
		} catch (Throwable e) {
			throw new RuntimeException(e);
		}
	}

	public PFGLDeviceRef createDevice(PFGLVersion glVersion) {
		try {
			return new PFGLDeviceRef((MemorySegment) MH_PFGLDeviceCreate.invokeExact(glVersion.value(), 0));
		} catch (Throwable e) {
			throw new RuntimeException(e);
		}
	}

	public PFResourceLoaderRef createEmbeddedResourceLoader() {
		try {
			return new PFResourceLoaderRef((MemorySegment) MH_PFEmbeddedResourceLoaderCreate.invokeExact());
		} catch (Throwable e) {
			throw new RuntimeException(e);
		}
	}

	public PFResourceLoaderRef createFilesystemResourceLoaderLocate() {
		try {
			return new PFResourceLoaderRef((MemorySegment) MH_PFFilesystemResourceLoaderLocate.invokeExact());
		} catch (Throwable e) {
			throw new RuntimeException(e);
		}
	}

	public PFResourceLoaderRef createFilesystemResourceLoaderFromPath(Path path) {
		try (var arena = Arena.ofConfined()) {
			var pathPtr = arena.allocateFrom(path.toString());
			return new PFResourceLoaderRef((MemorySegment) MH_PFFilesystemResourceLoaderFromPath.invokeExact(pathPtr));
		} catch (Throwable e) {
			throw new RuntimeException(e);
		}
	}

	public PFGLRendererRef createRenderer(PFGLDeviceRef device, PFResourceLoaderRef resourceLoader,
										  PFRendererMode mode, PFGLDestFramebufferRef framebuffer) {
		try (var arena = Arena.ofConfined()) {
			var options = arena.allocate(ML_PFRendererOptions);
			options.set(ML_PFDestFramebufferRef, ML_PFRendererOptions.byteOffset(groupElement(0)), framebuffer.ptr());
			options.set(JAVA_FLOAT, ML_PFRendererOptions.byteOffset(groupElement(1), groupElement("r")), 1f);
			options.set(JAVA_FLOAT, ML_PFRendererOptions.byteOffset(groupElement(1), groupElement("g")), 1f);
			options.set(JAVA_FLOAT, ML_PFRendererOptions.byteOffset(groupElement(1), groupElement("b")), 1f);
			options.set(JAVA_FLOAT, ML_PFRendererOptions.byteOffset(groupElement(1), groupElement("a")), 1f);
			options.set(JAVA_BYTE, ML_PFRendererOptions.byteOffset(groupElement(2)), (byte) 0);
			return new PFGLRendererRef((MemorySegment) MH_PFGLRendererCreate.invokeExact(device.ptr(),
					resourceLoader.ptr(),
					arena.allocateFrom(ML_PFRendererMode, mode.value),
//					arena.allocateFrom(ADDRESS, options)));
					options));
		} catch (Throwable e) {
			throw new RuntimeException(e);
		}
	}

	public PFCanvasFontContextRef createSystemFontContext() {
		try {
			return new PFCanvasFontContextRef((MemorySegment) MH_PFCanvasFontContextCreateWithSystemSource.invokeExact());
		} catch (Throwable e) {
			throw new RuntimeException(e);
		}
	}

	public PFCanvasRef createCanvas(PFCanvasFontContextRef fontContext, float width, float height) {
		try (var arena = Arena.ofConfined()) {
			var size = arena.allocateFrom(JAVA_FLOAT, width, height)
					.asSlice(0, ML_PFVector2F);
			return new PFCanvasRef((MemorySegment) MH_PFCanvasCreate.invokeExact(fontContext.ptr(), size));
		} catch (Throwable e) {
			throw new RuntimeException(e);
		}
	}

	public PFBuildOptionsRef createBuildOptions() {
		try {
			return new PFBuildOptionsRef((MemorySegment) MH_PFBuildOptionsCreate.invokeExact());
		} catch (Throwable e) {
			throw new RuntimeException(e);
		}
	}
}
