package dev.pfaff.altargui.render;

import dev.pfaff.altargui.GuiGlobals;
import dev.pfaff.altargui.internal.util.DumpUtil;
import dev.pfaff.glass.opengl.enums.GLBufferAccess;
import dev.pfaff.glass.opengl.enums.GLBufferTarget;
import dev.pfaff.glass.opengl.enums.GLBufferUsage;

import java.lang.foreign.Arena;
import java.lang.foreign.MemorySegment;

import static dev.pfaff.altargui.GuiGlobals.LOGGER;
import static dev.pfaff.altargui.render.GLFlags.ALWAYS_UNBIND;

public final class MappedBuffer extends AbstractMappedBuffer {
	private final int id;
	// @Stable
	private boolean initialized;

	public MappedBuffer(GLBufferTarget bindTarget, boolean isShared) {
		super(bindTarget, isShared);
		id = GLUtil.glGenBuffers();
	}

	@Override
	public void growBuffer(int capacity, Runnable onRecreate) {
		var newCapacity = this.capacity == 0 ? 2 : this.capacity;
		while (newCapacity < capacity) {
			newCapacity *= 2;
		}

		this.capacity = newCapacity;

		if (!this.initialized) {
			this.initialized = true;

			bind();
			try {
				onRecreate.run();
			} finally {
				unbind();
			}
		}
	}

	@Override
	protected void resetNoGrow() {
	}

	@Override
	public void beginUploading() {
		super.beginUploading();
		int capacity = this.capacity;
		if (mapArena != null) mapArena.close();
		// new arena, invalidate the old
		mapArena = newArena();
		bind();
		GLUtil.glBufferData(bindTarget, capacity, GLBufferUsage.STREAM_DRAW);
		//instanceDef.bindAttribs();
		// unmapping is automatic.
		mappedBuffer = GLUtil.glMapBuffer(bindTarget, GLBufferAccess.WRITE_ONLY, capacity)
				.reinterpret(mapArena, ignored -> {});
	}

	@Override
	public void finishUploading() {
		GLUtil.glUnmapBuffer(bindTarget);
		unbind();
		super.finishUploading();
	}

	public int id() {
		return id;
	}

	public void dump(String label, boolean capacity) {
		try (var arena = Arena.ofConfined()) {
			GuiGlobals.GL.glBindBuffer(GLBufferTarget.COPY_READ_BUFFER, id);
			MemorySegment buffer = GLUtil.glMapBuffer(GLBufferTarget.COPY_READ_BUFFER,
														 GLBufferAccess.READ_ONLY,
														 capacity ? this.capacity : length())
									 .reinterpret(arena, ignored -> {
										 GLUtil.glUnmapBuffer(GLBufferTarget.COPY_READ_BUFFER);
										 if (ALWAYS_UNBIND)
											 GuiGlobals.GL.glBindBuffer(GLBufferTarget.COPY_READ_BUFFER, 0);
									 });
			LOGGER.log(() -> STR."\{label}:  size: \{length()} bytes, capacity: \{capacity} bytes, buffer size: \{buffer.byteSize()} bytes");
			DumpUtil.dumpMemory(buffer);
		}
	}
}
