package dev.pfaff.altargui.render;

import dev.pfaff.altargui.GuiGlobals;
import dev.pfaff.altargui.render.objects.StaticGPUBuffer;
import dev.pfaff.glass.opengl.enums.GLBufferAccess;
import dev.pfaff.glass.opengl.enums.GLBufferTarget;

import java.lang.foreign.MemorySegment;
import java.nio.ByteBuffer;
import java.nio.ByteOrder;

import static dev.pfaff.altargui.render.GLFlags.ALWAYS_UNBIND;

public interface Sink {
	void checkRemaining(long length);

	void accept(byte value);
	void accept(short value);
	void accept(char value);
	void accept(int value);
	void accept(ByteOrder order, int value);
	void accept(long value);
	void accept(float value);
	void accept(double value);

	void accept(MemorySegment ms);

	void accept(ByteBuffer buffer, int offset, int length);

	default void accept(byte[] values, int offset, int length) {
		accept(MemorySegment.ofArray(values).asSlice(offset, length));
	}
	default void accept(short[] values, int offset, int length) {
		accept(MemorySegment.ofArray(values)
							.asSlice(offset * (long) Short.BYTES, length * (long) Short.BYTES));
	}
	default void accept(char[] values, int offset, int length) {
		accept(MemorySegment.ofArray(values)
							.asSlice(offset * (long) Character.BYTES, length * (long) Character.BYTES));
	}
	default void accept(int[] values, int offset, int length) {
		accept(MemorySegment.ofArray(values)
							.asSlice(offset * (long) Short.BYTES, length * (long) Short.BYTES));
	}
	default void accept(long[] values, int offset, int length) {
		accept(MemorySegment.ofArray(values)
							.asSlice(offset * (long) Long.BYTES, length * (long) Long.BYTES));
	}
	default void accept(float[] values, int offset, int length) {
		accept(MemorySegment.ofArray(values)
							.asSlice(offset * (long) Float.BYTES, length * (long) Float.BYTES));
	}
	default void accept(double[] values, int offset, int length) {
		accept(MemorySegment.ofArray(values)
							.asSlice(offset * (long) Double.BYTES, length * (long) Double.BYTES));
	}

	default void accept(byte[] values) {
		accept(MemorySegment.ofArray(values));
	}
	default void accept(short[] values) {
		accept(MemorySegment.ofArray(values));
	}
	default void accept(char[] values) {
		accept(MemorySegment.ofArray(values));
	}
	default void accept(int[] values) {
		accept(MemorySegment.ofArray(values));
	}
	default void accept(long[] values) {
		accept(MemorySegment.ofArray(values));
	}
	default void accept(float[] values) {
		accept(MemorySegment.ofArray(values));
	}
	default void accept(double[] values) {
		accept(MemorySegment.ofArray(values));
	}

	default void accept(ByteBuffer buffer) {
		int pos = buffer.position();
		accept(buffer, pos, buffer.limit() - pos);
	}

	default void accept(StaticGPUBuffer buffer) {
		GuiGlobals.GL.glBindBuffer(GLBufferTarget.COPY_READ_BUFFER, buffer.id());
		var ms = GLUtil.glMapBuffer(GLBufferTarget.COPY_READ_BUFFER, GLBufferAccess.READ_ONLY, buffer.size());
		try {
			accept(ms);
//		GL.glCopyBufferSubData(GL31.GL_COPY_READ_BUFFER,
//							   bindTarget,
//							   0,
//							   length,
//							   (long) buffer.stride() * buffer.count());
		} finally {
			GLUtil.glUnmapBuffer(GLBufferTarget.COPY_READ_BUFFER);
			if (ALWAYS_UNBIND) GuiGlobals.GL.glBindBuffer(GLBufferTarget.COPY_READ_BUFFER, 0);
		}

	}
}
