package dev.pfaff.altargui.render.objects;

import dev.pfaff.altargui.CanvasImpl;
import dev.pfaff.altargui.render.LayeredRenderBuffer;
import dev.pfaff.altargui.render.Sink;
import dev.pfaff.altargui.render.ViaIntermediateBuffer;
import dev.pfaff.altargui.types.Color;

import java.lang.foreign.MemoryLayout;

import static dev.pfaff.altargui.render.BufferUtil.emitShaderColor;
import static dev.pfaff.altargui.render.BufferUtil.emitVec2;
import static dev.pfaff.altargui.shader.Programs.RECT_INSTANCE_SPEC;
import static dev.pfaff.altargui.shader.Programs.rectRenderLayer;
import static java.util.FormatProcessor.FMT;

public final class RectRenderObject implements IRectRenderObject {
	public static final MemoryLayout LAYOUT = RECT_INSTANCE_SPEC.layout();

	private float x, y, width, height;
	/**
	 * RGBA
	 */
	private int color;

	public void updatePosition(float x, float y) {
		this.x = x;
		this.y = y;
	}

	public void updateSize(float width, float height) {
		this.width = width;
		this.height = height;
	}

	/**
	 * @param color RGBA
	 */
	public void updateColor(Color color) {
		this.color = color.rgba();
	}

	@Override
	public MemoryLayout instanceLayout() {
		return LAYOUT;
	}

	@Override
	public int size() {
		return IRectRenderObject.super.size();
	}

	@Override
	public void uploadTo(Sink sink) {
		sink.checkRemaining(LAYOUT.byteSize());
		emitVec2(sink, x, y);
		emitVec2(sink, width, height);
		emitShaderColor(sink, Color.fromRgba(color));
	}

	@Override
	public void uSpecUploadAtUnchecked(ViaIntermediateBuffer buffer, long offset) {
		IRectRenderObject.super.uSpecUploadAtUnchecked(buffer, offset);
	}

	@Override
	public void uSpecUploadCanvas(CanvasImpl canvas) {
		IRectRenderObject.super.uSpecUploadCanvas(canvas);
	}

	@Override
	public void uSpecUploadLayered(LayeredRenderBuffer buffer) {
		IRectRenderObject.super.uSpecUploadLayered(buffer);
	}

	@Override
	public int program() {
		return rectRenderLayer();
	}

	@Override
	public void close() {
	}

	@Override
	public String toString() {
		return FMT."RectRenderObject[x=\{x}, y=\{y}, width=\{width}, height=\{height}, color=%08x\{color}]";
	}
}
