package dev.pfaff.altargui.render.objects;

import dev.pfaff.altargui.CanvasImpl;
import dev.pfaff.altargui.render.*;

import java.lang.foreign.MemoryLayout;
import java.nio.ByteBuffer;
import java.nio.ByteOrder;
import java.util.function.Consumer;

public record StaticCPUBuffer(int program, byte[] buffer, MemoryLayout instanceLayout, int instanceCount) implements RenderObject {
	public StaticCPUBuffer {}

	public static StaticCPUBuffer build(int layer, MemoryLayout instanceLayout, int instanceCount, Consumer<? super ByteBufferSink> builder) {
		var buffer = new byte[(int)instanceLayout.byteSize() * instanceCount];

		var byteBuffer = ByteBuffer.wrap(buffer).order(ByteOrder.nativeOrder());
		builder.accept(new ByteBufferSink(byteBuffer));

		if (byteBuffer.position() != buffer.length) {
			throw new RuntimeException(STR."Buffer size not equal to allocated size: expected \{buffer.length}, found \{byteBuffer.position()}");
		}

		return new StaticCPUBuffer(layer, buffer, instanceLayout, instanceCount);
	}

	@Override
	public void uploadTo(Sink sink) {
		sink.accept(buffer);
	}

	@Override
	public void uSpecUploadAtUnchecked(ViaIntermediateBuffer buffer, long offset) {
		RenderObject.super.uSpecUploadAtUnchecked(buffer, offset);
	}

	@Override
	public void uSpecUploadCanvas(CanvasImpl canvas) {
		RenderObject.super.uSpecUploadCanvas(canvas);
	}

	@Override
	public void uSpecUploadLayered(LayeredRenderBuffer buffer) {
		RenderObject.super.uSpecUploadLayered(buffer);
	}

	@Override
	public void close() {}
}
