package dev.pfaff.altargui.render.objects;

import dev.pfaff.altargui.CanvasImpl;
import dev.pfaff.altargui.render.LayeredRenderBuffer;
import dev.pfaff.altargui.render.RenderObject;
import dev.pfaff.altargui.render.Sink;
import dev.pfaff.altargui.render.ViaIntermediateBuffer;
import dev.pfaff.altargui.types.Color;

import java.lang.foreign.MemoryLayout;

import static dev.pfaff.altargui.shader.Programs.textRenderLayer;
import static dev.pfaff.altargui.render.TextRenderLayerUtil.*;

public final class TextCharRenderObject implements RenderObject {
	private float x, y;
	private byte character;
	private float size = DEFAULT_SIZE;
	/**
	 * RGBA
	 */
	private int color = DEFAULT_COLOR.rgba();
	private int flags = DEFAULT_STYLE;

	public void updatePosition(float x, float y) {
		this.x = x;
		this.y = y;
	}

	public void updateCharacter(byte character) {
		this.character = character;
	}

	public void updateSize(float size) {
		this.size = size;
	}

	public void updateColor(Color color) {
		this.color = color.rgba();
	}

	public void updateStyle(int flags) {
		this.flags = flags;
	}

	@Override
	public int instanceCount() {
		return 1;
	}

	@Override
	public MemoryLayout instanceLayout() {
		return TEXT_CHAR_LAYOUT;
	}

	@Override
	public void uploadTo(Sink sink) {
		emitTextInstance(sink, x, y, size, character, Color.fromRgba(color), flags, 8f / 256f, 0f);
	}

	@Override
	public void uSpecUploadAtUnchecked(ViaIntermediateBuffer buffer, long offset) {
		RenderObject.super.uSpecUploadAtUnchecked(buffer, offset);
	}

	@Override
	public void uSpecUploadCanvas(CanvasImpl canvas) {
		RenderObject.super.uSpecUploadCanvas(canvas);
	}

	@Override
	public void uSpecUploadLayered(LayeredRenderBuffer buffer) {
		RenderObject.super.uSpecUploadLayered(buffer);
	}

	@Override
	public int program() {
		return textRenderLayer();
	}

	@Override
	public void close() {
	}
}
