package dev.pfaff.altargui.render.objects;

import dev.pfaff.altargui.CanvasImpl;
import dev.pfaff.altargui.types.Tex2d;
import dev.pfaff.altargui.texture.TextureHandle;
import dev.pfaff.altargui.render.LayeredRenderBuffer;
import dev.pfaff.altargui.render.Sink;
import dev.pfaff.altargui.render.ViaIntermediateBuffer;
import dev.pfaff.altargui.types.Color;

import java.lang.foreign.MemoryLayout;

import static dev.pfaff.altargui.render.BufferUtil.emitShaderColor;
import static dev.pfaff.altargui.render.BufferUtil.emitVec2;
import static dev.pfaff.altargui.shader.Programs.TEXTURED_RECT_INSTANCE_SPEC;
import static dev.pfaff.altargui.shader.Programs.texturedRectRenderLayer;
import static java.util.FormatProcessor.FMT;

public final class TexturedRectRenderObject implements IRectRenderObject {
	private float x, y, width, height;
	private float u, v, uExtent, vExtent;
	/**
	 * RGBA
	 */
	private int color;

	public void updatePosition(float x, float y) {
		this.x = x;
		this.y = y;
	}

	public void updateSize(float width, float height) {
		this.width = width;
		this.height = height;
	}

	/**
	 * Ignores texture id.
	 */
	public void updateTex(TextureHandle handle, Tex2d tex) {
		float u1 = tex.u1();
		float v1 = tex.v1();
		float u2 = tex.u2();
		float v2 = tex.v2();
		u1 *= handle.scaleFactorU;
		v1 *= handle.scaleFactorV;
		u2 *= handle.scaleFactorU;
		v2 *= handle.scaleFactorV;
		this.u = handle.u + u1;
		this.v = handle.v + v1;
		this.uExtent = u2 - u1;
		this.vExtent = v2 - v1;
	}

	/**
	 * @param color RGBA
	 */
	public void updateColor(Color color) {
		this.color = color.rgba();
	}

	@Override
	public MemoryLayout instanceLayout() {
		return TEXTURED_RECT_INSTANCE_SPEC.layout();
	}

	@Override
	public int size() {
		return IRectRenderObject.super.size();
	}

	@Override
	public void uploadTo(Sink sink) {
		emitVec2(sink, x, y);
		emitVec2(sink, width, height);
//		emitVec2AsNormUShort(sink, u, v);
//		emitVec2AsNormUShort(sink, uExtent, vExtent);
		emitVec2(sink, u, v);
		emitVec2(sink, uExtent, vExtent);
		emitShaderColor(sink, Color.fromRgba(color));
	}

	@Override
	public void uSpecUploadAtUnchecked(ViaIntermediateBuffer buffer, long offset) {
		IRectRenderObject.super.uSpecUploadAtUnchecked(buffer, offset);
	}

	@Override
	public void uSpecUploadCanvas(CanvasImpl canvas) {
		IRectRenderObject.super.uSpecUploadCanvas(canvas);
	}

	@Override
	public void uSpecUploadLayered(LayeredRenderBuffer buffer) {
		IRectRenderObject.super.uSpecUploadLayered(buffer);
	}

	@Override
	public int program() {
		return texturedRectRenderLayer();
	}

	@Override
	public void close() {
	}

	@Override
	public String toString() {
		return FMT."TexturedRectRenderObject[x=\{x}, y=\{y}, width=\{width}, height=\{height}, u=\{u}, v=\{v}, uExtent=\{uExtent}, vExtent=\{vExtent}, color=%08x\{color}]";
	}
}
