package dev.pfaff.altargui.resource;

import org.jspecify.annotations.NonNull;

import java.io.FileNotFoundException;
import java.io.IOException;
import java.io.InputStream;
import java.util.Objects;

/**
 * When running with modules, the {@code clazz}'s {@linkplain Class#getModule() module} must open the resource to
 * this class's module.
 *
 * @param clazz
 * @param resource
 */
public record ClassResourceLoadableResource(Class<?> clazz, String resource) implements LoadableResource.ByInputStream {
	public ClassResourceLoadableResource {
		Objects.requireNonNull(clazz);
		Objects.requireNonNull(resource);
		if (!resource.startsWith("/")) {
			throw new IllegalArgumentException("resource path must start with '/'");
		}
	}

	@Override
	public @NonNull InputStream openInputStream() throws IOException {
		var stream = this.clazz.getResourceAsStream(this.resource);
		if (stream == null) throw new FileNotFoundException(STR."No such resource \{this.resource} for \{this.clazz}");
		return stream;
	}
}
