package dev.pfaff.altargui.resource;

import it.unimi.dsi.fastutil.io.FastByteArrayInputStream;
import org.jspecify.annotations.NonNull;

import java.io.IOException;
import java.io.InputStream;
import java.lang.foreign.Arena;
import java.lang.foreign.MemorySegment;
import java.nio.charset.Charset;

import static java.lang.foreign.ValueLayout.JAVA_BYTE;

public interface LoadableResource {
	/**
	 * @implSpec The returned {@link MemorySegment} must be {@linkplain MemorySegment#isNative() native}.
	 */
	@NonNull
	MemorySegment readBytes() throws IOException;

	@NonNull
	default String readString(Charset charset) throws IOException {
		var ms = readBytes();
		if (ms.heapBase().orElse(null) instanceof byte[] bytes) {
			return new String(bytes, (int) ms.address(), (int) ms.byteSize(), charset);
		}
		return new String(readBytes().toArray(JAVA_BYTE), charset);
	}

	@NonNull
	InputStream openInputStream() throws IOException;

	interface ByBytes extends LoadableResource {
		@Override
		@NonNull
		default InputStream openInputStream() throws IOException {
			// TODO: optimized InputStream implementation
			return new FastByteArrayInputStream(readBytes().toArray(JAVA_BYTE));
		}
	}

	interface ByInputStream extends LoadableResource {
		@Override
		@NonNull
		default MemorySegment readBytes() throws IOException {
			try (var is = openInputStream()) {
				return Arena.ofAuto().allocateFrom(JAVA_BYTE, is.readAllBytes());
			}
		}
	}
}
