package dev.pfaff.altargui.resource;

import org.jspecify.annotations.NonNull;

import java.io.FileNotFoundException;
import java.io.IOException;
import java.io.InputStream;
import java.util.Objects;

/**
 * When running with modules, the {@code module} must open the resource to this class's module.
 *
 * @param module the module owning the resource
 * @param resource the path to the resource within the module
 */
public record ModuleResourceLoadableResource(Module module, String resource) implements LoadableResource.ByInputStream {
	public ModuleResourceLoadableResource {
		Objects.requireNonNull(module);
		Objects.requireNonNull(resource);
		if (!resource.startsWith("/")) {
			throw new IllegalArgumentException("resource path must start with '/'");
		}
	}

	@Override
	public @NonNull InputStream openInputStream() throws IOException {
		var stream = this.module.getResourceAsStream(this.resource);
		if (stream == null) throw new FileNotFoundException(STR."No such resource \{this.resource} for \{this.module}");
		return stream;
	}
}
