package dev.pfaff.altargui.widgets;

import dev.pfaff.altargui.*;
import dev.pfaff.altargui.handlers.MousePressHandler;
import dev.pfaff.altargui.handlers.MouseReleaseHandler;
import dev.pfaff.altargui.types.Color;
import org.jetbrains.annotations.NotNull;

import java.util.BitSet;
import java.util.function.IntConsumer;

public record ButtonWidget(Widget child, IntConsumer onClick) implements SingleChildWidget {
	private static final Widget INLINE_STYLE_BG_NORMAL = RectWidget.TRANSPARENT;
	private static final Widget INLINE_STYLE_BG_HOVERED = RectWidget.colored(Color.fromArgb(0x90_808080));

	private static final Widget INLINE_STYLE_BG = new MouseOverBuilder(hovered -> {
		return hovered ? INLINE_STYLE_BG_HOVERED : INLINE_STYLE_BG_NORMAL;
	});

	public ButtonWidget(Widget child, Runnable onClick) {
		this(child, btn -> {
			if (btn == 0) onClick.run();
		});
	}

	public static ButtonWidget inlineStyle(Widget child, IntConsumer onClick) {
		return new ButtonWidget(child.withBackground(INLINE_STYLE_BG), onClick);
	}

	public static ButtonWidget inlineStyle(Widget child, Runnable onClick) {
		return new ButtonWidget(child.withBackground(INLINE_STYLE_BG), onClick);
	}

	@Override
	public @NotNull AbstractState<ButtonWidget> instantiate(WidgetInstantiationArgs args) {
		return new State(args);
	}

	private static final class State extends AbstractState<ButtonWidget> implements MousePressHandler, MouseReleaseHandler {
		private IntConsumer onClick;
		private final BitSet buttonsPressed = new BitSet(3);

		private State(WidgetInstantiationArgs args) {
			super(args);
		}

		@Override
		public void configure(StateContext stateContext, ButtonWidget widget, boolean init) {
			super.configure(stateContext, widget, init);
			this.onClick = widget.onClick;
		}

		@Override
		public boolean mousePressed(double mouseX, double mouseY, int button) {
			if (this.isInside(mouseX, mouseY)) {
				this.buttonsPressed.set(button);
				return true;
			}
			return false;
		}

		@Override
		public boolean mouseReleased(double mouseX, double mouseY, int button) {
			if (this.buttonsPressed.get(button)) {
				if (this.isInside(mouseX, mouseY)) {
					this.onClick.accept(button);
					return true;
				}
				this.buttonsPressed.clear(button);
			}
			return false;
		}
	}
}
