package dev.pfaff.altargui.widgets;

import dev.pfaff.altargui.*;
import dev.pfaff.altargui.inspector.DebugLayer;
import dev.pfaff.altargui.internal.util.StencilLayoutContext;
import org.jetbrains.annotations.NotNull;

import java.util.Objects;

import static dev.pfaff.altargui.util.ClipUtil.clipExtent;

/**
 * Clips the child to the size allocated by the parent.
 */
public record ClipWidget(Widget child) implements SingleChildWidget {
	public static final DebugLayer DEBUG_LAYER_PRE_UNSCISSORED = new DebugLayer(ClipWidget.class, "pre_unscissored");
	public static final DebugLayer DEBUG_LAYER_PRE_SCISSORED = new DebugLayer(ClipWidget.class, "pre_scissored");
	public static final DebugLayer DEBUG_LAYER_POST_UNSCISSORED = new DebugLayer(ClipWidget.class, "post_unscissored");

	public ClipWidget {
		Objects.requireNonNull(child);
	}

	@Override
	public @NotNull AbstractState<ClipWidget> instantiate(WidgetInstantiationArgs args) {
		return new State(args);
	}

	private static final class State extends AbstractState<ClipWidget> {
		private State(WidgetInstantiationArgs args) {
			super(args);
		}

		@Override
		protected void computeStencilLayout(StateContext context) {
			final var allocWidth = this.allocatedWidth();
			final var allocHeight = this.allocatedHeight();

			var childElement = getChildElement();
			StencilLayoutContext.updateMountAndLayout(context, this, childElement, (e, ctx) -> {
				e.setAvailableSize(allocWidth, allocHeight);
				e.stencilLayout(ctx);
			});

			setComputedSize(Math.min(allocWidth, childElement.computedWidth()),
					Math.min(allocHeight, childElement.computedHeight()));
		}

		@Override
		protected void computeScissorLayout() {
			var clippedWidth = clipExtent(this.clippedX(),
					this.clippedWidth(),
					this.allocatedX(),
					this.computedWidth());
			var clippedHeight = clipExtent(this.clippedY(),
					this.clippedHeight(),
					this.allocatedY(),
					this.computedHeight());
			var childElement = this.getChildElement();
			childElement.setClippedRect(clippedWidth.start(), clippedHeight.start(), clippedWidth.extent(), clippedHeight.extent());
			childElement.scissorLayout();
		}

		@Override
		protected void computeMount(MountContext context) {
			context.pushScissor(this);
			super.computeMount(context);
			context.popScissor();
		}

		@Override
		public void uSpecMount(MountContext context) {
			super.uSpecMount(context);
		}
	}
}
