package dev.pfaff.altargui.widgets;

import dev.pfaff.altargui.StateContext;
import dev.pfaff.altargui.Widget;
import dev.pfaff.altargui.WidgetInstance;
import dev.pfaff.altargui.WidgetInstantiationArgs;
import dev.pfaff.altargui.element.*;
import org.jetbrains.annotations.NotNull;

public record DraggableWidget(Widget child, int initialX, int initialY) implements Widget {
	@Override
	public @NotNull WidgetInstance instantiate(WidgetInstantiationArgs args) {
		return new State(args);
	}

	private static final class State extends StatefulWidgetInstance<DraggableWidget> {
		private Widget child;
		private float x;
		private float y;

		private final DragTarget.DragHandler dragHandler = (dragging, mouseX, mouseY, deltaX, deltaY) -> {
			if (dragging && (deltaX != 0 || deltaY != 0)) {
				x += deltaX;
				y += deltaY;
				markNeedsRebuild();
			}
		};

		private State(WidgetInstantiationArgs args) {
			super(args);
		}

		@Override
		public void configure(StateContext stateContext, DraggableWidget widget, boolean init) {
			if (init) {
				x = widget.initialX;
				y = widget.initialY;
			}
			if (init || child != widget.child) {
				child = widget.child;
				markNeedsRebuild();
			}
		}

		@Override
		protected @NotNull Widget build() {
			return new DragTarget(child, dragHandler).atAbsolutePosition(x, y);
		}
	}
}
