package dev.pfaff.altargui.widgets;

import dev.pfaff.altargui.StateContext;
import dev.pfaff.altargui.Widget;
import dev.pfaff.altargui.WidgetInstance;
import dev.pfaff.altargui.WidgetInstantiationArgs;
import dev.pfaff.altargui.selector.BuiltinSelector;
import dev.pfaff.altargui.selector.Children;
import dev.pfaff.altargui.selector.NthElement;
import org.jetbrains.annotations.NotNull;
import org.jspecify.annotations.NonNull;
import org.jspecify.annotations.Nullable;

import java.util.Objects;
import java.util.function.Consumer;

public record IdWidget(@NonNull String id, @NotNull Widget child) implements SingleChildWidget {
	public IdWidget {
		Objects.requireNonNull(id, "id cannot be null");
		Objects.requireNonNull(child, "child cannot be null");
	}

	@Override
	public @NotNull AbstractState<IdWidget> instantiate(WidgetInstantiationArgs args) {
		return new State(args);
	}

	public static final class State extends AbstractState<IdWidget> {
		@Nullable
		private String id;

		private State(WidgetInstantiationArgs args) {
			super(args);
		}

		@NonNull
		public String id() {
			//noinspection DataFlowIssue
			return this.id;
		}

		@Override
		public void configure(StateContext stateContext, IdWidget widget, boolean init) {
			super.configure(stateContext, widget, init);
			this.id = widget.id;
		}

		@Override
		public void applyBuiltinSelector(BuiltinSelector selector, Consumer<WidgetInstance> consumer) {
			// this widget intentionally forwards Children and NthElement selectors to its child.
			switch (selector) {
				case Children ignored -> BuiltinSelector.applyNext(this, selector, consumer);
				case NthElement ignored -> BuiltinSelector.applyNext(this, selector, consumer);
			}
		}
	}
}
