package dev.pfaff.altargui.widgets;

import dev.pfaff.altargui.StateContext;
import dev.pfaff.altargui.Widget;
import dev.pfaff.altargui.WidgetInstance;
import dev.pfaff.altargui.WidgetInstantiationArgs;
import dev.pfaff.altargui.element.SingleChildWidgetInstance;
import org.jetbrains.annotations.NotNull;

/**
 * A widget that calls the provided {@code builder} every layout.
 */
public record LayoutBuilderWidget(Builder builder) implements Widget {
	@Override
	public @NotNull WidgetInstance<LayoutBuilderWidget> instantiate(WidgetInstantiationArgs args) {
		return new State(args);
	}

	@FunctionalInterface
	public interface Builder {
		Widget build(float width, float height);
	}

	private static final class State extends SingleChildWidgetInstance<LayoutBuilderWidget> {
		private Builder builder;

		private State(WidgetInstantiationArgs args) {
			super(args);
		}

		@Override
		public void configure(StateContext stateContext, LayoutBuilderWidget widget, boolean init) {
			if (!widget.builder.equals(this.builder)) {
				this.builder = widget.builder;
				markNeedsStencilLayout();
			}
		}

		@Override
		protected void computeStencilLayout(StateContext context) {
			// TODO: add a profiling counter for the number of computeLayout calls with unchanged allocated rect
			// TODO: cache result of builder.build if profiling counter suggests it would be beneficial
			setChild(context, this.builder.build(allocatedWidth(), allocatedHeight()), false);
			super.computeStencilLayout(context);
		}
	}
}
