package dev.pfaff.altargui.widgets;

import dev.pfaff.altargui.StateContext;
import dev.pfaff.altargui.Widget;
import dev.pfaff.altargui.WidgetInstance;
import dev.pfaff.altargui.WidgetInstantiationArgs;
import dev.pfaff.altargui.element.StatefulWidgetInstance;
import dev.pfaff.altargui.function.BooleanFunction;
import dev.pfaff.altargui.handlers.MouseMoveHandler;
import org.jetbrains.annotations.NotNull;

import java.util.Objects;

public record MouseOverBuilder(BooleanFunction<Widget> builder) implements Widget {
	public MouseOverBuilder {
		Objects.requireNonNull(builder);
	}

	public MouseOverBuilder(Widget childWhenHovered, Widget childWhenNotHovered) {
		Objects.requireNonNull(childWhenHovered);
		Objects.requireNonNull(childWhenNotHovered);
		this(hovered -> hovered ? childWhenHovered : childWhenNotHovered);
	}

	@Override
	public @NotNull WidgetInstance<MouseOverBuilder> instantiate(WidgetInstantiationArgs args) {
		return new State(args);
	}

	private static final class State extends StatefulWidgetInstance<MouseOverBuilder>
			implements MouseMoveHandler {
		private BooleanFunction<Widget> builder;
		private boolean hovered;

		private State(WidgetInstantiationArgs args) {
			super(args);
		}

		@Override
		public void configure(StateContext stateContext, MouseOverBuilder widget, boolean init) {
			if (!widget.builder.equals(this.builder)) {
				this.builder = widget.builder;
				markNeedsRebuild();
			}
		}

		@Override
		protected @NotNull Widget build() {
			return this.builder.apply(this.hovered);
		}

		@Override
		protected void computeStencilLayout(StateContext context) {
			super.computeStencilLayout(context);
			if (!this.needsTraceLayout.get()) {
				this.mouseMoved(guiContext().lastMouseX(), guiContext().lastMouseY());
			}
		}

		@Override
		protected void computeTraceLayout() {
			super.computeTraceLayout();
			this.mouseMoved(guiContext().lastMouseX(), guiContext().lastMouseY());
		}

		@Override
		public boolean mouseMoved(double mouseX, double mouseY) {
			return mouseMoved((float) mouseX, (float) mouseY);
		}

		private boolean mouseMoved(float mouseX, float mouseY) {
			var hovered = isInside(mouseX, mouseY);
			if (this.hovered != hovered) {
				this.hovered = hovered;
				markNeedsRebuild();
				return hovered;
			}
			return false;
		}
	}
}
