package dev.pfaff.altargui.widgets;

import dev.pfaff.altargui.StateContext;
import dev.pfaff.altargui.Widget;
import dev.pfaff.altargui.WidgetInstance;
import dev.pfaff.altargui.WidgetInstantiationArgs;
import dev.pfaff.altargui.types.Alignment;
import org.jetbrains.annotations.NotNull;

import java.util.Objects;

public record MultiChildAttributesWidget(Widget child, boolean tightX, boolean tightY, Alignment alignmentX, Alignment alignmentY) implements AttributesWidget {
	public MultiChildAttributesWidget {
		Objects.requireNonNull(child);
		Objects.requireNonNull(alignmentX);
		Objects.requireNonNull(alignmentY);
	}

	public MultiChildAttributesWidget(Widget child) {
		this(child, false, false, Alignment.Start, Alignment.Start);
	}

	@Override
	public MultiChildAttributesWidget withMultiChildAttributes() {
		return this;
	}

	public MultiChildAttributesWidget tighten() {
		return new MultiChildAttributesWidget(child, true, true, alignmentX, alignmentY);
	}

	public MultiChildAttributesWidget tightenWidth() {
		return new MultiChildAttributesWidget(child, true, tightY, alignmentX, alignmentY);
	}

	public MultiChildAttributesWidget tightenHeight() {
		return new MultiChildAttributesWidget(child, tightX, true, alignmentX, alignmentY);
	}

	public MultiChildAttributesWidget tightAlign(Alignment alignmentX, Alignment alignmentY) {
		return new MultiChildAttributesWidget(child, tightX, tightY, alignmentX, alignmentY);
	}

	public MultiChildAttributesWidget tightAlign(Alignment alignment) {
		return tightAlign(alignment, alignment);
	}

	public MultiChildAttributesWidget tightAlignX(Alignment alignmentX) {
		return tightAlign(alignmentX, alignmentY);
	}

	public MultiChildAttributesWidget tightAlignY(Alignment alignmentY) {
		return tightAlign(alignmentX, alignmentY);
	}

	@Override
	public @NotNull State instantiate(WidgetInstantiationArgs args) {
		return new State(args);
	}

	public static final class State extends AbstractState<MultiChildAttributesWidget> {
		private boolean tightX;
		private boolean tightY;
		private Alignment alignmentX = Alignment.Start;
		private Alignment alignmentY = Alignment.Start;

		private State(WidgetInstantiationArgs args) {
			super(args);
		}

		public boolean isTightX() {
			return this.tightX;
		}

		public boolean isTightY() {
			return this.tightY;
		}

		public Alignment getAlignmentX() {
			return this.alignmentX;
		}

		public Alignment getAlignmentY() {
			return this.alignmentY;
		}

		@Override
		public void configure(StateContext stateContext, MultiChildAttributesWidget widget, boolean init) {
			super.configure(stateContext, widget, init);

			boolean stencilDirty = false, traceDirty = false;

			if (this.tightX != widget.tightX) {
				this.tightX = widget.tightX;
				stencilDirty = true;
			}
			if (this.tightY != widget.tightY) {
				this.tightY = widget.tightY;
				stencilDirty = true;
			}

			if (this.alignmentX != widget.alignmentX) {
				this.alignmentX = widget.alignmentX;
				traceDirty = true;
			}
			if (this.alignmentY != widget.alignmentY) {
				this.alignmentY = widget.alignmentY;
				traceDirty = true;
			}

			if (stencilDirty) markParentNeedsStencilLayout();
			if (traceDirty) markParentNeedsTraceLayout();
		}

		private void markParentNeedsStencilLayout() {
			if (parent() instanceof WidgetInstance<?> parentWidgetInstance) {
				parentWidgetInstance.needsStencilLayout.set();
			}
		}

		private void markParentNeedsTraceLayout() {
			if (parent() instanceof WidgetInstance<?> parentWidgetInstance) {
				parentWidgetInstance.needsTraceLayout.set();
			}
		}
	}
}
