package dev.pfaff.altargui.widgets;

import dev.pfaff.altargui.StateContext;
import dev.pfaff.altargui.Widget;
import dev.pfaff.altargui.WidgetInstantiationArgs;
import dev.pfaff.altargui.element.SingleChildWidgetInstance;
import org.jetbrains.annotations.MustBeInvokedByOverriders;
import org.jetbrains.annotations.NotNull;

public interface SingleChildWidget extends Widget {
    Widget child();

    @Override
    default @NotNull AbstractState<? extends SingleChildWidget> instantiate(WidgetInstantiationArgs args) {
        return new AutoLayoutState<>(args);
    }

    static abstract class AbstractState<W extends SingleChildWidget> extends SingleChildWidgetInstance<W> {
        protected AbstractState(WidgetInstantiationArgs args) {
            super(args);
        }

        @Override
        @MustBeInvokedByOverriders
        public void configure(StateContext stateContext, W widget, boolean init) {
            setChild(stateContext, widget.child());
        }
    }

    static final class AutoLayoutState<W extends SingleChildWidget> extends AbstractState<W> {
        private AutoLayoutState(WidgetInstantiationArgs args) {
            super(args);
        }
    }
}
