package dev.pfaff.altargui.widgets;

import dev.pfaff.altargui.StateContext;
import dev.pfaff.altargui.Widget;
import dev.pfaff.altargui.WidgetInstance;
import dev.pfaff.altargui.WidgetInstantiationArgs;
import dev.pfaff.altargui.element.StatefulWidgetInstance;
import dev.pfaff.altargui.types.Color;
import it.unimi.dsi.fastutil.floats.FloatConsumer;
import org.jetbrains.annotations.NotNull;

import java.util.List;

import static dev.pfaff.altargui.widgets.MultiChildWidget.stack;

public record SliderWidget(float value, FloatConsumer onChanged) implements Widget {
	public SliderWidget {
		if (value < 0f || value > 1f) throw new IllegalArgumentException("Invalid slider value");
	}

	@Override
	public @NotNull WidgetInstance instantiate(WidgetInstantiationArgs args) {
		return new State(args);
	}

	private static final class State extends StatefulWidgetInstance<SliderWidget> {
		private float value;
		private FloatConsumer onChanged;

		private State(WidgetInstantiationArgs args) {
			super(args);
		}

		@Override
		public void configure(StateContext stateContext, SliderWidget widget, boolean init) {
			onChanged = widget.onChanged;
			if (value != widget.value) {
				value = widget.value;
				markNeedsRebuild();
			}
		}

		@Override
		protected @NotNull Widget build() {
			return stack(List.of(
					RectWidget.GENERIC_HEAVY_HUD_BACKGROUND
							.withVerticalMargin(2f),
					new DragTarget(RectWidget.colored(Color.fromArgb(0xff_4444ff))
											 .withAbsoluteWidth(10)
											 .atRelativePosition(-5 + (value * allocatedWidth()), 0),
								   (dragging, __, ___, deltaX, deltaY) -> {
									   onChanged.accept(value = Math.clamp(value + (deltaX / allocatedWidth()), 0f, 1f));
									   markNeedsRebuild();
								   })
			))
					.withAbsoluteHeight(10);
		}
	}
}
